from typing import List

from qtpy.QtCore import Qt
from qtpy.QtGui import QFont, QFontMetrics, QTextLayout


class _GenericEliding:
    """A mixin to provide capabilities to elide text (could add '…') to fit width."""

    _elide_mode: Qt.TextElideMode = Qt.TextElideMode.ElideRight
    _text: str = ""
    # the 2 is a magic number that prevents the ellipses from going missing
    # in certain cases (?)
    _ellipses_width: int = 2

    # Public methods

    def elideMode(self) -> Qt.TextElideMode:
        """The current Qt.TextElideMode."""
        return self._elide_mode

    def setElideMode(self, mode: Qt.TextElideMode) -> None:
        """Set the elide mode to a Qt.TextElideMode."""
        self._elide_mode = Qt.TextElideMode(mode)

    def full_text(self) -> str:
        """The current text without eliding."""
        return self._text

    def setEllipsesWidth(self, width: int) -> None:
        """A width value to take into account ellipses width when eliding text.

        The value is deducted from the widget width when computing the elided version
        of the text.
        """
        self._ellipses_width = width

    @staticmethod
    def wrapText(text, width, font=None) -> List[str]:
        """Returns `text`, split as it would be wrapped for `width`, given `font`.

        Static method.
        """
        tl = QTextLayout(text, font or QFont())
        tl.beginLayout()
        lines = []
        while True:
            ln = tl.createLine()
            if not ln.isValid():
                break
            ln.setLineWidth(width)
            start = ln.textStart()
            lines.append(text[start : start + ln.textLength()])
        tl.endLayout()
        return lines

    # private implementation methods

    def _elidedText(self) -> str:
        """Return `self._text` elided to `width`."""
        fm = QFontMetrics(self.font())
        ellipses_width = 0
        if self._elide_mode != Qt.TextElideMode.ElideNone:
            ellipses_width = self._ellipses_width
        width = self.width() - ellipses_width
        if not getattr(self, "wordWrap", None) or not self.wordWrap():
            return fm.elidedText(self._text, self._elide_mode, width)

        # get number of lines we can fit without eliding
        nlines = self.height() // fm.height() - 1
        # get the last line (elided)
        text = self._wrappedText()
        last_line = fm.elidedText("".join(text[nlines:]), self._elide_mode, width)
        # join them
        return "".join(text[:nlines] + [last_line])

    def _wrappedText(self) -> List[str]:
        return _GenericEliding.wrapText(self._text, self.width(), self.font())



from qtpy.QtCore import QSize, Qt
from qtpy.QtGui import QResizeEvent

from superqt import QElidingLineEdit

TEXT = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do"
ELLIPSIS = "…"


def test_init_text_eliding_line_edit(qtbot):
    wdg = QElidingLineEdit(TEXT)
    qtbot.addWidget(wdg)
    oldsize = QSize(100, 20)
    wdg.resize(oldsize)
    assert wdg._elidedText().endswith(ELLIPSIS)
    newsize = QSize(500, 20)
    wdg.resize(newsize)
    wdg.resizeEvent(QResizeEvent(oldsize, newsize))  # for test coverage
    assert wdg._elidedText() == TEXT
    assert wdg.text() == TEXT


def test_set_text_eliding_line_edit(qtbot):
    wdg = QElidingLineEdit()
    qtbot.addWidget(wdg)
    wdg.resize(500, 20)
    wdg.setText(TEXT)
    assert not wdg._elidedText().endswith(ELLIPSIS)
    wdg.resize(100, 20)
    assert wdg._elidedText().endswith(ELLIPSIS)


def test_set_elide_mode_eliding_line_edit(qtbot):
    wdg = QElidingLineEdit()
    qtbot.addWidget(wdg)
    wdg.resize(500, 20)
    wdg.setText(TEXT)
    assert not wdg._elidedText().endswith(ELLIPSIS)
    wdg.resize(100, 20)
    # ellipses should be to the right
    assert wdg._elidedText().endswith(ELLIPSIS)

    # ellipses should be to the left
    wdg.setElideMode(Qt.TextElideMode.ElideLeft)
    assert wdg._elidedText().startswith(ELLIPSIS)
    assert wdg.elideMode() == Qt.TextElideMode.ElideLeft

    # no ellipses should be shown
    wdg.setElideMode(Qt.TextElideMode.ElideNone)
    assert ELLIPSIS not in wdg._elidedText()


def test_set_elipses_width_eliding_line_edit(qtbot):
    wdg = QElidingLineEdit()
    qtbot.addWidget(wdg)
    wdg.resize(500, 20)
    wdg.setText(TEXT)
    assert not wdg._elidedText().endswith(ELLIPSIS)
    wdg.setEllipsesWidth(int(wdg.width() / 2))
    assert wdg._elidedText().endswith(ELLIPSIS)



from __future__ import annotations

import inspect
import time
import warnings
from functools import partial, wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Generator,
    Generic,
    Sequence,
    TypeVar,
    overload,
)

from qtpy.QtCore import QObject, QRunnable, QThread, QThreadPool, QTimer, Signal

if TYPE_CHECKING:
    _T = TypeVar("_T")

    class SigInst(Generic[_T]):
        @staticmethod
        def connect(slot: Callable[[_T], Any], type: type | None = ...) -> None:
            ...

        @staticmethod
        def disconnect(slot: Callable[[_T], Any] = ...) -> None:
            ...

        @staticmethod
        def emit(*args: _T) -> None:
            ...

    from typing_extensions import Literal, ParamSpec

    _P = ParamSpec("_P")
# maintain runtime compatibility with older typing_extensions
else:
    try:
        from typing_extensions import ParamSpec

        _P = ParamSpec("_P")
    except ImportError:
        _P = TypeVar("_P")

_Y = TypeVar("_Y")
_S = TypeVar("_S")
_R = TypeVar("_R")


def as_generator_function(
    func: Callable[_P, _R]
) -> Callable[_P, Generator[None, None, _R]]:
    """Turns a regular function (single return) into a generator function."""

    @wraps(func)
    def genwrapper(*args: Any, **kwargs: Any) -> Generator[None, None, _R]:
        yield
        return func(*args, **kwargs)

    return genwrapper


class WorkerBaseSignals(QObject):
    started = Signal()  # emitted when the work is started
    finished = Signal()  # emitted when the work is finished
    _finished = Signal(object)  # emitted when the work is finished to delete
    returned = Signal(object)  # emitted with return value
    errored = Signal(object)  # emitted with error object on Exception
    warned = Signal(tuple)  # emitted with showwarning args on warning


class WorkerBase(QRunnable, Generic[_R]):
    """Base class for creating a Worker that can run in another thread.

    Parameters
    ----------
    SignalsClass : type, optional
        A QObject subclass that contains signals, by default WorkerBaseSignals

    Attributes
    ----------
    signals: WorkerBaseSignals
        signal emitter object. To allow identify which worker thread emitted signal.
    """

    #: A set of Workers.  Add to set using `WorkerBase.start`
    _worker_set: ClassVar[set[WorkerBase]] = set()
    returned: SigInst[_R]
    errored: SigInst[Exception]
    warned: SigInst[tuple]
    started: SigInst[None]
    finished: SigInst[None]

    def __init__(
        self,
        func: Callable[_P, _R] | None = None,
        SignalsClass: type[WorkerBaseSignals] = WorkerBaseSignals,
    ) -> None:
        super().__init__()
        self._abort_requested = False
        self._running = False
        self.signals = SignalsClass()

    def __getattr__(self, name: str) -> SigInst:
        """Pass through attr requests to signals to simplify connection API.

        The goal is to enable `worker.yielded.connect` instead of
        `worker.signals.yielded.connect`. Because multiple inheritance of Qt
        classes is not well supported in PyQt, we have to use composition here
        (signals are provided by QObjects, and QRunnable is not a QObject). So
        this passthrough allows us to connect to signals on the `_signals`
        object.
        """
        # the Signal object is actually a class attribute
        attr = getattr(self.signals.__class__, name, None)
        if isinstance(attr, Signal):
            # but what we need to connect to is the instantiated signal
            # (which is of type `SignalInstance` in PySide and
            # `pyqtBoundSignal` in PyQt)
            return getattr(self.signals, name)
        raise AttributeError(
            f"{self.__class__.__name__!r} object has no attribute {name!r}"
        )

    def quit(self) -> None:
        """Send a request to abort the worker.

        !!! note
            It is entirely up to subclasses to honor this method by checking
            `self.abort_requested` periodically in their `worker.work`
            method, and exiting if `True`.
        """
        self._abort_requested = True

    @property
    def abort_requested(self) -> bool:
        """Whether the worker has been requested to stop."""
        return self._abort_requested

    @property
    def is_running(self) -> bool:
        """Whether the worker has been started."""
        return self._running

    def run(self) -> None:
        """Start the worker.

        The end-user should never need to call this function.
        But it cannot be made private or renamed, since it is called by Qt.

        The order of method calls when starting a worker is:

        ```
           calls QThreadPool.globalInstance().start(worker)
           |               triggered by the QThreadPool.start() method
           |               |             called by worker.run
           |               |             |
           V               V             V
           worker.start -> worker.run -> worker.work
        ```

        **This** is the function that actually gets called when calling
        `QThreadPool.start(worker)`.  It simply wraps the `work()`
        method, and emits a few signals.  Subclasses should NOT override this
        method (except with good reason), and instead should implement
        `work()`.
        """
        self.started.emit()
        self._running = True
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings("always")
                warnings.showwarning = lambda *w: self.warned.emit(w)
                result = self.work()
            if isinstance(result, Exception):
                if isinstance(result, RuntimeError):
                    # The Worker object has likely been deleted.
                    # A deleted wrapped C/C++ object may result in a runtime
                    # error that will cause segfault if we try to do much other
                    # than simply notify the user.
                    warnings.warn(
                        f"RuntimeError in aborted thread: {result}",
                        RuntimeWarning,
                        stacklevel=2,
                    )
                    return
                else:
                    raise result
            if not self.abort_requested:
                self.returned.emit(result)
        except Exception as exc:
            self.errored.emit(exc)
        self._running = False
        self.finished.emit()
        self._finished.emit(self)

    def work(self) -> Exception | _R:
        """Main method to execute the worker.

        The end-user should never need to call this function.
        But subclasses must implement this method (See
        [`GeneratorFunction.work`][superqt.utils._qthreading.GeneratorWorker.work] for
        an example implementation). Minimally, it should check `self.abort_requested`
        periodically and exit if True.

        Examples
        --------
        ```python
        class MyWorker(WorkerBase):

            def work(self):
                i = 0
                while True:
                    if self.abort_requested:
                        self.aborted.emit()
                        break
                    i += 1
                    if i > max_iters:
                        break
                    time.sleep(0.5)
        ```
        """
        raise NotImplementedError(
            f'"{self.__class__.__name__}" failed to define work() method'
        )

    def start(self) -> None:
        """Start this worker in a thread and add it to the global threadpool.

        The order of method calls when starting a worker is:

        ```
           calls QThreadPool.globalInstance().start(worker)
           |               triggered by the QThreadPool.start() method
           |               |             called by worker.run
           |               |             |
           V               V             V
           worker.start -> worker.run -> worker.work
        ```
        """
        if self in self._worker_set:
            raise RuntimeError("This worker is already started!")

        # This will raise a RunTimeError if the worker is already deleted
        repr(self)

        self._worker_set.add(self)
        self._finished.connect(self._set_discard)
        if QThread.currentThread().loopLevel():
            # if we're in a thread with an eventloop, queue the worker to start
            start_ = partial(QThreadPool.globalInstance().start, self)
            QTimer.singleShot(1, start_)
        else:
            # otherwise start it immediately
            QThreadPool.globalInstance().start(self)

    @classmethod
    def _set_discard(cls, obj: WorkerBase) -> None:
        cls._worker_set.discard(obj)

    @classmethod
    def await_workers(cls, msecs: int | None = None) -> None:
        """Ask all workers to quit, and wait up to `msec` for quit.

        Attempts to clean up all running workers by calling `worker.quit()`
        method.  Any workers in the `WorkerBase._worker_set` set will have this
        method.

        By default, this function will block indefinitely, until worker threads
        finish.  If a timeout is provided, a `RuntimeError` will be raised if
        the workers do not gracefully exit in the time requests, but the threads
        will NOT be killed.  It is (currently) left to the user to use their OS
        to force-quit rogue threads.

        !!! important

            If the user does not put any yields in their function, and the function
            is super long, it will just hang... For instance, there's no graceful
            way to kill this thread in python:

            ```python
            @thread_worker
            def ZZZzzz():
                time.sleep(10000000)
            ```

            This is why it's always advisable to use a generator that periodically
            yields for long-running computations in another thread.

            See [this stack-overflow
            post](https://stackoverflow.com/questions/323972/is-there-any-way-to-kill-a-thread)
            for a good discussion on the difficulty of killing a rogue python thread:

        Parameters
        ----------
        msecs : int, optional
            Waits up to msecs milliseconds for all threads to exit and removes all
            threads from the thread pool. If msecs is `None` (the default), the
            timeout is ignored (waits for the last thread to exit).

        Raises
        ------
        RuntimeError
            If a timeout is provided and workers do not quit successfully within
            the time allotted.
        """
        for worker in cls._worker_set:
            worker.quit()

        msecs = msecs if msecs is not None else -1
        if not QThreadPool.globalInstance().waitForDone(msecs):
            raise RuntimeError(
                f"Workers did not quit gracefully in the time allotted ({msecs} ms)"
            )


class FunctionWorker(WorkerBase[_R]):
    """QRunnable with signals that wraps a simple long-running function.

    !!! note
        `FunctionWorker` does not provide a way to stop a very long-running
        function (e.g. `time.sleep(10000)`).  So whenever possible, it is better to
        implement your long running function as a generator that yields periodically,
        and use the [`GeneratorWorker`][superqt.utils.GeneratorWorker] instead.

    Parameters
    ----------
    func : Callable
        A function to call in another thread
    *args
        will be passed to the function
    **kwargs
        will be passed to the function

    Raises
    ------
    TypeError
        If `func` is a generator function and not a regular function.
    """

    def __init__(self, func: Callable[_P, _R], *args, **kwargs):
        if inspect.isgeneratorfunction(func):
            raise TypeError(
                f"Generator function {func} cannot be used with FunctionWorker, "
                "use GeneratorWorker instead",
            )
        super().__init__()

        self._func = func
        self._args = args
        self._kwargs = kwargs

    def work(self) -> _R:
        return self._func(*self._args, **self._kwargs)


class GeneratorWorkerSignals(WorkerBaseSignals):
    yielded = Signal(object)  # emitted with yielded values (if generator used)
    paused = Signal()  # emitted when a running job has successfully paused
    resumed = Signal()  # emitted when a paused job has successfully resumed
    aborted = Signal()  # emitted when a running job is successfully aborted


class GeneratorWorker(WorkerBase, Generic[_Y, _S, _R]):
    """QRunnable with signals that wraps a long-running generator.

    Provides a convenient way to run a generator function in another thread,
    while allowing 2-way communication between threads, using plain-python
    generator syntax in the original function.

    Parameters
    ----------
    func : callable
        The function being run in another thread.  May be a generator function.
    SignalsClass : type, optional
        A QObject subclass that contains signals, by default
        GeneratorWorkerSignals
    *args
        Will be passed to func on instantiation
    **kwargs
        Will be passed to func on instantiation
    """

    yielded: SigInst[_Y]
    paused: SigInst[None]
    resumed: SigInst[None]
    aborted: SigInst[None]

    def __init__(
        self,
        func: Callable[_P, Generator[_Y, _S | None, _R]],
        *args,
        SignalsClass: type[WorkerBaseSignals] = GeneratorWorkerSignals,
        **kwargs,
    ):
        if not inspect.isgeneratorfunction(func):
            raise TypeError(
                f"Regular function {func} cannot be used with GeneratorWorker, "
                "use FunctionWorker instead",
            )
        super().__init__(SignalsClass=SignalsClass)

        self._gen = func(*args, **kwargs)
        self._incoming_value: _S | None = None
        self._pause_requested = False
        self._resume_requested = False
        self._paused = False

        # polling interval: ONLY relevant if the user paused a running worker
        self._pause_interval = 0.01
        self.pbar = None

    def work(self) -> _R | None | Exception:
        """Core event loop that calls the original function.

        Enters a continual loop, yielding and returning from the original
        function.  Checks for various events (quit, pause, resume, etc...).
        (To clarify: we are creating a rudimentary event loop here because
        there IS NO Qt event loop running in the other thread to hook into)
        """
        while True:
            if self.abort_requested:
                self.aborted.emit()
                break
            if self._paused:
                if self._resume_requested:
                    self._paused = False
                    self._resume_requested = False
                    self.resumed.emit()
                else:
                    time.sleep(self._pause_interval)
                    continue
            elif self._pause_requested:
                self._paused = True
                self._pause_requested = False
                self.paused.emit()
                continue
            try:
                _input = self._next_value()
                output = self._gen.send(_input)
                self.yielded.emit(output)
            except StopIteration as exc:
                return exc.value
            except RuntimeError as exc:
                # The worker has probably been deleted.  warning will be
                # emitted in `WorkerBase.run`
                return exc
        return None

    def send(self, value: _S):
        """Send a value into the function (if a generator was used)."""
        self._incoming_value = value

    def _next_value(self) -> _S | None:
        out = None
        if self._incoming_value is not None:
            out = self._incoming_value
            self._incoming_value = None
        return out

    @property
    def is_paused(self) -> bool:
        """Whether the worker is currently paused."""
        return self._paused

    def toggle_pause(self) -> None:
        """Request to pause the worker if playing or resume if paused."""
        if self.is_paused:
            self._resume_requested = True
        else:
            self._pause_requested = True

    def pause(self) -> None:
        """Request to pause the worker."""
        if not self.is_paused:
            self._pause_requested = True

    def resume(self) -> None:
        """Send a request to resume the worker."""
        if self.is_paused:
            self._resume_requested = True


#############################################################################

# convenience functions for creating Worker instances


@overload
def create_worker(
    func: Callable[_P, Generator[_Y, _S, _R]],
    *args,
    _start_thread: bool | None = None,
    _connect: dict[str, Callable | Sequence[Callable]] | None = None,
    _worker_class: type[GeneratorWorker] | type[FunctionWorker] | None = None,
    _ignore_errors: bool = False,
    **kwargs,
) -> GeneratorWorker[_Y, _S, _R]:
    ...


@overload
def create_worker(
    func: Callable[_P, _R],
    *args,
    _start_thread: bool | None = None,
    _connect: dict[str, Callable | Sequence[Callable]] | None = None,
    _worker_class: type[GeneratorWorker] | type[FunctionWorker] | None = None,
    _ignore_errors: bool = False,
    **kwargs,
) -> FunctionWorker[_R]:
    ...


def create_worker(
    func: Callable,
    *args,
    _start_thread: bool | None = None,
    _connect: dict[str, Callable | Sequence[Callable]] | None = None,
    _worker_class: type[GeneratorWorker] | type[FunctionWorker] | None = None,
    _ignore_errors: bool = False,
    **kwargs,
) -> FunctionWorker | GeneratorWorker:
    """Convenience function to start a function in another thread.

    By default, uses `FunctionWorker` for functions and `GeneratorWorker` for
    generators, but a custom `WorkerBase` subclass may be provided.  If so, it must be a
    subclass of `WorkerBase`, which defines a standard set of signals and a run method.

    Parameters
    ----------
    func : Callable
        The function to call in another thread.
    _start_thread : bool
        Whether to immediaetly start the thread.  If False, the returned worker
        must be manually started with `worker.start()`. by default it will be
        `False` if the `_connect` argument is `None`, otherwise `True`.
    _connect : Dict[str, Union[Callable, Sequence]], optional
        A mapping of `"signal_name"` -> `callable` or list of `callable`:
        callback functions to connect to the various signals offered by the
        worker class. by default `None`
    _worker_class : type of `GeneratorWorker` or `FunctionWorker`, optional
        The [`WorkerBase`][superqt.utils.WorkerBase] to instantiate, by default
        [`FunctionWorker`][superqt.utils.FunctionWorker] will be used if `func` is a
        regular function, and [`GeneratorWorker`][superqt.utils.GeneratorWorker] will be
        used if it is a generator.
    _ignore_errors : bool
        If `False` (the default), errors raised in the other thread will be
        reraised in the main thread (makes debugging significantly easier).
    *args
        will be passed to `func`
    **kwargs
        will be passed to `func`

    Returns
    -------
    worker : WorkerBase
        An instantiated worker.  If `_start_thread` was `False`, the worker
        will have a `.start()` method that can be used to start the thread.

    Raises
    ------
    TypeError
        If a worker_class is provided that is not a subclass of WorkerBase.
    TypeError
        If _connect is provided and is not a dict of `{str: callable}`

    Examples
    --------
    ```python
    def long_function(duration):
        import time
        time.sleep(duration)

    worker = create_worker(long_function, 10)
    ```
    """
    worker: FunctionWorker | GeneratorWorker

    if not _worker_class:
        if inspect.isgeneratorfunction(func):
            _worker_class = GeneratorWorker
        else:
            _worker_class = FunctionWorker

    if not inspect.isclass(_worker_class) and issubclass(_worker_class, WorkerBase):
        raise TypeError(f"Worker {_worker_class} must be a subclass of WorkerBase")

    worker = _worker_class(func, *args, **kwargs)

    if _connect is not None:
        if not isinstance(_connect, dict):
            raise TypeError("The '_connect' argument must be a dict")

        if _start_thread is None:
            _start_thread = True

        for key, val in _connect.items():
            _val = val if isinstance(val, (tuple, list)) else [val]
            for v in _val:
                if not callable(v):
                    raise TypeError(
                        f"_connect[{key!r}] must be a function or sequence of functions"
                    )
                getattr(worker, key).connect(v)

    # if the user has not provided a default connection for the "errored"
    # signal... and they have not explicitly set `ignore_errors=True`
    # Then rereaise any errors from the thread.
    if not _ignore_errors and not (_connect or {}).get("errored", False):

        def reraise(e):
            raise e

        worker.errored.connect(reraise)

    if _start_thread:
        worker.start()
    return worker


@overload
def thread_worker(
    function: Callable[_P, Generator[_Y, _S, _R]],
    start_thread: bool | None = None,
    connect: dict[str, Callable | Sequence[Callable]] | None = None,
    worker_class: type[WorkerBase] | None = None,
    ignore_errors: bool = False,
) -> Callable[_P, GeneratorWorker[_Y, _S, _R]]:
    ...


@overload
def thread_worker(
    function: Callable[_P, _R],
    start_thread: bool | None = None,
    connect: dict[str, Callable | Sequence[Callable]] | None = None,
    worker_class: type[WorkerBase] | None = None,
    ignore_errors: bool = False,
) -> Callable[_P, FunctionWorker[_R]]:
    ...


@overload
def thread_worker(
    function: Literal[None] = None,
    start_thread: bool | None = None,
    connect: dict[str, Callable | Sequence[Callable]] | None = None,
    worker_class: type[WorkerBase] | None = None,
    ignore_errors: bool = False,
) -> Callable[[Callable], Callable[_P, FunctionWorker | GeneratorWorker]]:
    ...


def thread_worker(
    function: Callable | None = None,
    start_thread: bool | None = None,
    connect: dict[str, Callable | Sequence[Callable]] | None = None,
    worker_class: type[WorkerBase] | None = None,
    ignore_errors: bool = False,
):
    """Decorator that runs a function in a separate thread when called.

    When called, the decorated function returns a
    [`WorkerBase`][superqt.utils.WorkerBase].  See
    [`create_worker`][superqt.utils.create_worker] for additional keyword arguments that
    can be used
    when calling the function.

    The returned worker will have these signals:

    - **started**: emitted when the work is started
    - **finished**: emitted when the work is finished
    - **returned**: emitted with return value
    - **errored**: emitted with error object on Exception

    It will also have a `worker.start()` method that can be used to start
    execution of the function in another thread. (useful if you need to connect
    callbacks to signals prior to execution)

    If the decorated function is a generator, the returned worker will also
    provide these signals:

    - **yielded**: emitted with yielded values
    - **paused**: emitted when a running job has successfully paused
    - **resumed**: emitted when a paused job has successfully resumed
    - **aborted**: emitted when a running job is successfully aborted

    And these methods:

    - **quit**: ask the thread to quit
    - **toggle_paused**: toggle the running state of the thread.
    - **send**: send a value into the generator.  (This requires that your
      decorator function uses the `value = yield` syntax)

    Parameters
    ----------
    function : callable
        Function to call in another thread.  For communication between threads
        may be a generator function.
    start_thread : bool
        Whether to immediaetly start the thread.  If False, the returned worker
        must be manually started with `worker.start()`. by default it will be
        `False` if the `_connect` argument is `None`, otherwise `True`.
    connect : Dict[str, Union[Callable, Sequence]]
        A mapping of `"signal_name"` -> `callable` or list of `callable`:
        callback functions to connect to the various signals offered by the
        worker class. by default None
    worker_class : Type[WorkerBase]
        The [`WorkerBase`][superqt.utils.WorkerBase] to instantiate, by default
        [`FunctionWorker`][superqt.utils.FunctionWorker] will be used if `func` is a
        regular function, and [`GeneratorWorker`][superqt.utils.GeneratorWorker] will be
        used if it is a generator.
    ignore_errors : bool
        If `False` (the default), errors raised in the other thread will be
        reraised in the main thread (makes debugging significantly easier).

    Returns
    -------
    callable
        function that creates a worker, puts it in a new thread and returns
        the worker instance.

    Examples
    --------
    ```python
    @thread_worker
    def long_function(start, end):
        # do work, periodically yielding
        i = start
        while i <= end:
            time.sleep(0.1)
            yield i

        # do teardown
        return 'anything'

    # call the function to start running in another thread.
    worker = long_function()

    # connect signals here if desired... or they may be added using the
    # `connect` argument in the `@thread_worker` decorator... in which
    # case the worker will start immediately when long_function() is called
    worker.start()
    ```
    """

    def _inner(func):
        @wraps(func)
        def worker_function(*args, **kwargs):
            # decorator kwargs can be overridden at call time by using the
            # underscore-prefixed version of the kwarg.
            kwargs["_start_thread"] = kwargs.get("_start_thread", start_thread)
            kwargs["_connect"] = kwargs.get("_connect", connect)
            kwargs["_worker_class"] = kwargs.get("_worker_class", worker_class)
            kwargs["_ignore_errors"] = kwargs.get("_ignore_errors", ignore_errors)
            return create_worker(
                func,
                *args,
                **kwargs,
            )

        return worker_function

    return _inner if function is None else _inner(function)


############################################################################

# This is a variant on the above pattern, it uses QThread instead of Qrunnable
# see https://doc.qt.io/qt-5/threads-technologies.html#comparison-of-solutions
# (it appears from that table that QRunnable cannot emit or receive signals,
# but we circumvent that here with our WorkerBase class that also inherits from
# QObject... providing signals/slots).
#
# A benefit of the QRunnable pattern is that Qt manages the threads for you,
# in the QThreadPool.globalInstance() ... making it easier to reuse threads,
# and reduce overhead.
#
# However, a disadvantage is that you have no access to (and therefore less
# control over) the QThread itself.  See for example all of the methods
# provided on the QThread object: https://doc.qt.io/qt-5/qthread.html

if TYPE_CHECKING:

    class WorkerProtocol(QObject):
        finished: Signal

        def work(self) -> None:
            ...


def new_worker_qthread(
    Worker: type[WorkerProtocol],
    *args,
    _start_thread: bool = False,
    _connect: dict[str, Callable] | None = None,
    **kwargs,
):
    """Convenience function to start a worker in a `QThread`.

    thread, not as the actual code or object that runs in that
    thread.  The QThread object is created on the main thread and lives there.

    Worker objects which derive from QObject are the things that actually do
    the work. They can be moved to a QThread as is done here.

    ??? "Mostly ignorable detail"

        While the signals/slots syntax of the worker looks very similar to
        standard "single-threaded" signals & slots, note that inter-thread
        signals and slots (automatically) use an event-based QueuedConnection, while
        intra-thread signals use a DirectConnection. See [Signals and Slots Across
        Threads](https://doc.qt.io/qt-5/threads-qobject.html#signals-and-slots-across-threads>)

    Parameters
    ----------
    Worker : QObject
        QObject type that implements a `work()` method.  The Worker should also
        emit a finished signal when the work is done.
    _start_thread : bool
        If True, thread will be started immediately, otherwise, thread must
        be manually started with thread.start().
    _connect : dict
        Optional dictionary of {signal: function} to connect to the new worker.
        for instance:  _connect = {'incremented': myfunc} will result in:
        worker.incremented.connect(myfunc)
    *args
        will be passed to the Worker class on instantiation.
    **kwargs
        will be passed to the Worker class on instantiation.

    Returns
    -------
    worker : WorkerBase
        The created worker.
    thread : QThread
        The thread on which the worker is running.

    Examples
    --------
    Create some QObject that has a long-running work method:

    ```python

    class Worker(QObject):

        finished = Signal()
        increment = Signal(int)

        def __init__(self, argument):
            super().__init__()
            self.argument = argument

        @Slot()
        def work(self):
            # some long running task...
            import time
            for i in range(10):
                time.sleep(1)
                self.increment.emit(i)
            self.finished.emit()

    worker, thread = new_worker_qthread(
        Worker,
        'argument',
        _start_thread=True,
        _connect={'increment': print},
    )
    ```
    """
    if _connect and not isinstance(_connect, dict):
        raise TypeError("_connect parameter must be a dict")

    thread = QThread()
    worker = Worker(*args, **kwargs)
    worker.moveToThread(thread)
    thread.started.connect(worker.work)
    worker.finished.connect(thread.quit)
    worker.finished.connect(worker.deleteLater)
    thread.finished.connect(thread.deleteLater)

    if _connect:
        [getattr(worker, key).connect(val) for key, val in _connect.items()]

    if _start_thread:
        thread.start()  # sometimes need to connect stuff before starting
    return worker, thread



from qtpy.QtCore import Qt
from qtpy.QtGui import QFocusEvent, QResizeEvent
from qtpy.QtWidgets import QLineEdit

from ._eliding import _GenericEliding


class QElidingLineEdit(_GenericEliding, QLineEdit):
    """A QLineEdit variant that will elide text (could add '…') to fit width.

    QElidingLineEdit()
    QElidingLineEdit(parent: Optional[QWidget])
    QElidingLineEdit(text: str, parent: Optional[QWidget] = None)

    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if args and isinstance(args[0], str):
            self.setText(args[0])
        # The `textEdited` signal doesn't trigger the `textChanged` signal if
        # text is changed with `setText`, so we connect to `textEdited` to only
        # update _text when text is being edited by the user graphically.
        self.textEdited.connect(self._update_text)

    # Reimplemented _GenericEliding methods

    def setElideMode(self, mode: Qt.TextElideMode) -> None:
        """Set the elide mode to a Qt.TextElideMode.

        The text shown is updated to the elided version only if the widget is not
        focused.
        """
        super().setElideMode(mode)
        if not self.hasFocus():
            super().setText(self._elidedText())

    def setEllipsesWidth(self, width: int) -> None:
        """A width value to take into account ellipses width when eliding text.

        The value is deducted from the widget width when computing the elided version
        of the text. The text shown is updated to the elided version only if the widget
        is not focused.
        """
        super().setEllipsesWidth(width)
        if not self.hasFocus():
            super().setText(self._elidedText())

    # Reimplemented QT methods

    def text(self) -> str:
        """Return the label's text being shown.

        If no text has been set this will return an empty string.
        """
        return self._text

    def setText(self, text) -> None:
        """Set the line edit's text.

        Setting the text clears any previous content.
        NOTE: we set the QLineEdit private text to the elided version
        """
        self._text = text
        if not self.hasFocus():
            super().setText(self._elidedText())

    def focusInEvent(self, event: QFocusEvent) -> None:
        """Set the full text when the widget is focused."""
        super().setText(self._text)
        super().focusInEvent(event)

    def focusOutEvent(self, event: QFocusEvent) -> None:
        """Set an elided version of the text (if needed) when the focus is out."""
        super().setText(self._elidedText())
        super().focusOutEvent(event)

    def resizeEvent(self, event: QResizeEvent) -> None:
        """Update elided text being shown when the widget is resized."""
        if not self.hasFocus():
            super().setText(self._elidedText())
        super().resizeEvent(event)

    # private implementation methods

    def _update_text(self, text: str) -> None:
        """Update only the actual text of the widget.

        The actual text is the text the widget has without eliding.
        """
        self._text = text



from superqt import QSearchableComboBox


class TestSearchableComboBox:
    def test_constructor(self, qtbot):
        widget = QSearchableComboBox()
        qtbot.addWidget(widget)

    def test_add_items(self, qtbot):
        widget = QSearchableComboBox()
        qtbot.addWidget(widget)
        widget.addItems(["foo", "bar"])
        assert widget.completer_object.model().rowCount() == 2
        widget.addItem("foobar")
        assert widget.completer_object.model().rowCount() == 3
        widget.insertItem(1, "baz")
        assert widget.completer_object.model().rowCount() == 4
        widget.insertItems(2, ["bazbar", "foobaz"])
        assert widget.completer_object.model().rowCount() == 6
        assert widget.itemText(0) == "foo"
        assert widget.itemText(1) == "baz"
        assert widget.itemText(2) == "bazbar"

    def test_completion(self, qtbot):
        widget = QSearchableComboBox()
        qtbot.addWidget(widget)
        widget.addItems(["foo", "bar", "foobar", "baz", "bazbar", "foobaz"])

        widget.completer_object.setCompletionPrefix("fo")
        assert widget.completer_object.completionCount() == 3
        assert widget.completer_object.currentCompletion() == "foo"
        widget.completer_object.setCurrentRow(1)
        assert widget.completer_object.currentCompletion() == "foobar"
        widget.completer_object.setCurrentRow(2)
        assert widget.completer_object.currentCompletion() == "foobaz"



import sys
from pathlib import Path

import pytest
from qtpy.QtGui import QIcon, QPixmap

from superqt.fonticon import _plugins, icon
from superqt.fonticon._qfont_icon import QFontIconStore

FIXTURES = Path(__file__).parent / "fixtures"


@pytest.fixture
def plugin_store(qapp, monkeypatch):
    _path = [str(FIXTURES), *sys.path.copy()]
    store = QFontIconStore().instance()
    with monkeypatch.context() as m:
        m.setattr(sys, "path", _path)
        yield store
    store.clear()


def test_plugin(plugin_store):
    assert not _plugins.loaded()
    icn = icon("ico.smiley")
    assert _plugins.loaded() == {"ico": ["smiley"]}
    assert isinstance(icn, QIcon)
    assert isinstance(icn.pixmap(40, 40), QPixmap)



from unittest.mock import Mock

from qtpy.QtCore import QObject, Signal

from superqt.utils import signals_blocked
from superqt.utils._util import get_max_args


def test_signal_blocker(qtbot):
    """make sure context manager signal blocker works"""

    class Emitter(QObject):
        sig = Signal()

    obj = Emitter()
    receiver = Mock()
    obj.sig.connect(receiver)

    # make sure signal works
    with qtbot.waitSignal(obj.sig):
        obj.sig.emit()

    receiver.assert_called_once()
    receiver.reset_mock()

    with signals_blocked(obj):
        obj.sig.emit()
        qtbot.wait(10)

    receiver.assert_not_called()


def test_get_max_args_simple():
    def fun1():
        pass

    assert get_max_args(fun1) == 0

    def fun2(a):
        pass

    assert get_max_args(fun2) == 1

    def fun3(a, b=1):
        pass

    assert get_max_args(fun3) == 2

    def fun4(a, *, b=2):
        pass

    assert get_max_args(fun4) == 1

    def fun5(a, *b):
        pass

    assert get_max_args(fun5) is None

    assert get_max_args(print) is None


def test_get_max_args_wrapped():
    from functools import partial, wraps

    def fun1(a, b):
        pass

    assert get_max_args(partial(fun1, 1)) == 1

    def dec(fun):
        @wraps(fun)
        def wrapper(*args, **kwargs):
            return fun(*args, **kwargs)

        return wrapper

    assert get_max_args(dec(fun1)) == 2


def test_get_max_args_methods():
    class A:
        def fun1(self):
            pass

        def fun2(self, a):
            pass

        def __call__(self, a, b=1):
            pass

    assert get_max_args(A().fun1) == 0
    assert get_max_args(A().fun2) == 1
    assert get_max_args(A()) == 2



import platform

from qtpy.QtCore import QSize, Qt
from qtpy.QtGui import QResizeEvent

from superqt import QElidingLabel

TEXT = (
    "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do "
    "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad "
    "minim ven iam, quis nostrud exercitation ullamco laborisnisi ut aliquip "
    "ex ea commodo consequat. Duis aute irure dolor inreprehenderit in voluptate "
    "velit esse cillum dolore eu fugiat nullapariatur."
)
ELLIPSIS = "…"


def test_eliding_label(qtbot):
    wdg = QElidingLabel(TEXT)
    qtbot.addWidget(wdg)
    assert wdg._elidedText().endswith(ELLIPSIS)
    oldsize = wdg.size()
    newsize = QSize(200, 20)
    wdg.resize(newsize)
    wdg.resizeEvent(QResizeEvent(oldsize, newsize))  # for test coverage
    assert wdg.text() == TEXT


def test_wrapped_eliding_label(qtbot):
    wdg = QElidingLabel(TEXT)
    qtbot.addWidget(wdg)
    assert not wdg.wordWrap()
    assert 630 < wdg.sizeHint().width() < 640
    assert wdg._elidedText().endswith(ELLIPSIS)
    wdg.resize(QSize(200, 100))
    assert wdg.text() == TEXT
    assert wdg._elidedText().endswith(ELLIPSIS)
    wdg.setWordWrap(True)
    assert wdg.wordWrap()
    assert wdg.text() == TEXT
    assert wdg._elidedText().endswith(ELLIPSIS)
    # just empirically from CI ... stupid
    if platform.system() == "Linux":
        assert wdg.sizeHint() in (QSize(200, 198), QSize(200, 154))
    elif platform.system() == "Windows":
        assert wdg.sizeHint() in (QSize(200, 160), QSize(200, 118))
    elif platform.system() == "Darwin":
        assert wdg.sizeHint() == QSize(200, 176)
        # TODO: figure out how to test these on all platforms on CI
        wdg.resize(wdg.sizeHint())
        assert wdg._elidedText() == TEXT


def test_shorter_eliding_label(qtbot):
    short = "asd a ads sd flksdf dsf lksfj sd lsdjf sd lsdfk sdlkfj s"
    wdg = QElidingLabel()
    qtbot.addWidget(wdg)
    wdg.setText(short)
    assert not wdg._elidedText().endswith(ELLIPSIS)
    wdg.resize(100, 20)
    assert wdg._elidedText().endswith(ELLIPSIS)
    wdg.setElideMode(Qt.TextElideMode.ElideLeft)
    assert wdg._elidedText().startswith(ELLIPSIS)
    assert wdg.elideMode() == Qt.TextElideMode.ElideLeft


def test_wrap_text():
    wrap = QElidingLabel.wrapText(TEXT, 200)
    assert isinstance(wrap, list)
    assert all(isinstance(x, str) for x in wrap)
    assert 9 <= len(wrap) <= 13



import math
import os

import pytest
from qtpy import API_NAME
from qtpy.QtWidgets import QStyleOptionSlider

from superqt import (
    QDoubleRangeSlider,
    QDoubleSlider,
    QLabeledDoubleRangeSlider,
    QLabeledDoubleSlider,
)

from ._testutil import _linspace

range_types = {QDoubleRangeSlider, QLabeledDoubleRangeSlider}


@pytest.fixture(
    params=[
        QDoubleSlider,
        QLabeledDoubleSlider,
        QDoubleRangeSlider,
        QLabeledDoubleRangeSlider,
    ]
)
def ds(qtbot, request):
    # convenience fixture that converts value() and setValue()
    # to let us use setValue((a, b)) for both range and non-range sliders
    cls = request.param
    wdg = cls()
    qtbot.addWidget(wdg)

    def assert_val_type():
        type_ = float
        if cls in range_types:
            assert all(isinstance(i, type_) for i in wdg.value())  # sourcery skip
        else:
            assert isinstance(wdg.value(), type_)

    def assert_val_eq(val):
        assert wdg.value() == val if cls is QDoubleRangeSlider else val[0]

    wdg.assert_val_type = assert_val_type
    wdg.assert_val_eq = assert_val_eq

    if cls not in range_types:
        superset = wdg.setValue

        def _safe_set(val):
            superset(val[0] if isinstance(val, tuple) else val)

        wdg.setValue = _safe_set

    return wdg


def test_double_sliders(ds):
    ds.setMinimum(10)
    ds.setMaximum(99)
    ds.setValue((20, 40))
    ds.setSingleStep(1)
    assert ds.minimum() == 10
    assert ds.maximum() == 99
    ds.assert_val_eq((20, 40))
    assert ds.singleStep() == 1

    ds.assert_val_eq((20, 40))
    ds.assert_val_type()

    ds.setValue((20.23, 40.23))
    ds.assert_val_eq((20.23, 40.23))
    ds.assert_val_type()

    assert ds.minimum() == 10
    assert ds.maximum() == 99
    assert ds.singleStep() == 1
    ds.assert_val_eq((20.23, 40.23))
    ds.setValue((20.2343, 40.2342))
    ds.assert_val_eq((20.2343, 40.2342))

    ds.assert_val_eq((20.2343, 40.2342))
    assert ds.minimum() == 10
    assert ds.maximum() == 99
    assert ds.singleStep() == 1

    ds.assert_val_eq((20.2343, 40.2342))
    assert ds.minimum() == 10
    assert ds.maximum() == 99
    assert ds.singleStep() == 1


def test_double_sliders_small(ds):
    ds.setMaximum(1)
    ds.setValue((0.5, 0.9))
    assert ds.minimum() == 0
    assert ds.maximum() == 1
    ds.assert_val_eq((0.5, 0.9))

    ds.setValue((0.122233, 0.72644353))
    ds.assert_val_eq((0.122233, 0.72644353))


def test_double_sliders_big(ds):
    ds.setValue((20, 80))
    ds.setMaximum(5e14)
    assert ds.minimum() == 0
    assert ds.maximum() == 5e14
    ds.setValue((1.74e9, 1.432e10))
    ds.assert_val_eq((1.74e9, 1.432e10))


@pytest.mark.skipif(
    os.name == "nt" and API_NAME == "PyQt6", reason="Not ready for pyqt6"
)
def test_signals(ds, qtbot):
    with qtbot.waitSignal(ds.valueChanged):
        ds.setValue((10, 20))

    with qtbot.waitSignal(ds.rangeChanged):
        ds.setMinimum(0.5)

    with qtbot.waitSignal(ds.rangeChanged):
        ds.setMaximum(3.7)

    with qtbot.waitSignal(ds.rangeChanged):
        ds.setRange(1.2, 3.3)


@pytest.mark.parametrize("mag", list(range(4, 37, 4)) + list(range(-4, -37, -4)))
def test_slider_extremes(mag, qtbot):
    sld = QDoubleSlider()
    _mag = 10**mag
    with qtbot.waitSignal(sld.rangeChanged):
        sld.setRange(-_mag, _mag)
    for i in _linspace(-_mag, _mag, 10):
        sld.setValue(i)
        assert math.isclose(sld.value(), i, rel_tol=1e-8)
        sld.initStyleOption(QStyleOptionSlider())



import logging

from qtpy import QtCore

from superqt import QMessageHandler


def test_message_handler():
    with QMessageHandler() as mh:
        QtCore.qDebug("debug")
        QtCore.qWarning("warning")
        QtCore.qCritical("critical")

    assert len(mh.records) == 3
    assert mh.records[0].level == logging.DEBUG
    assert mh.records[1].level == logging.WARNING
    assert mh.records[2].level == logging.CRITICAL

    assert "3 records" in repr(mh)


def test_message_handler_with_logger(caplog):
    logger = logging.getLogger("test_logger")
    caplog.set_level(logging.DEBUG, logger="test_logger")
    with QMessageHandler(logger):
        QtCore.qDebug("debug")
        QtCore.qWarning("warning")
        QtCore.qCritical("critical")

    assert len(caplog.records) == 3
    assert caplog.records[0].message == "debug"
    assert caplog.records[0].levelno == logging.DEBUG
    assert caplog.records[1].message == "warning"
    assert caplog.records[1].levelno == logging.WARNING
    assert caplog.records[2].message == "critical"
    assert caplog.records[2].levelno == logging.CRITICAL



import sys
import warnings
from importlib import abc, util

from qtpy import *  # noqa

warnings.warn(
    "The superqt.qtcompat module is deprecated as of v0.3.0. "
    "Please import from `qtpy` instead.",
    stacklevel=2,
)


# forward any requests for superqt.qtcompat.* to qtpy.*
class SuperQtImporter(abc.MetaPathFinder):
    def find_spec(self, fullname: str, path, target=None):  # type: ignore
        """Forward any requests for superqt.qtcompat.* to qtpy.*."""
        if fullname.startswith(__name__):
            return util.find_spec(fullname.replace(__name__, "qtpy"))


sys.meta_path.append(SuperQtImporter())



from pint import Quantity

from superqt import QQuantity


def test_qquantity(qtbot):
    w = QQuantity(1, "m")
    qtbot.addWidget(w)

    assert w.value() == 1 * w.unitRegistry().meter
    assert w.magnitude() == 1
    assert w.units() == w.unitRegistry().meter
    assert w.text() == "1 meter"
    w.setUnits("cm")
    assert w.value() == 100 * w.unitRegistry().centimeter
    assert w.magnitude() == 100
    assert w.units() == w.unitRegistry().centimeter
    assert w.text() == "100.0 centimeter"
    w.setMagnitude(10)
    assert w.value() == 10 * w.unitRegistry().centimeter
    assert w.magnitude() == 10
    assert w.units() == w.unitRegistry().centimeter
    assert w.text() == "10 centimeter"
    w.setValue(1 * w.unitRegistry().meter)
    assert w.value() == 1 * w.unitRegistry().meter
    assert w.magnitude() == 1
    assert w.units() == w.unitRegistry().meter
    assert w.text() == "1 meter"

    w.setUnits(None)
    assert w.isDimensionless()
    assert w.unitsComboBox().currentText() == "-----"
    assert w.magnitude() == 1


def test_change_qquantity_value(qtbot):
    w = QQuantity()
    qtbot.addWidget(w)
    assert w.value() == Quantity(0)
    w.setValue(Quantity("1 meter"))
    assert w.value() == Quantity("1 meter")



from superqt import QSearchableListWidget


class TestSearchableListWidget:
    def test_create(self, qtbot):
        widget = QSearchableListWidget()
        qtbot.addWidget(widget)
        widget.addItem("aaa")
        assert widget.count() == 1

    def test_add_items(self, qtbot):
        widget = QSearchableListWidget()
        qtbot.addWidget(widget)
        widget.addItems(["foo", "bar"])
        assert widget.count() == 2
        widget.insertItems(1, ["baz", "foobaz"])
        widget.insertItem(2, "foobar")
        assert widget.count() == 5
        assert widget.item(0).text() == "foo"
        assert widget.item(1).text() == "baz"
        assert widget.item(2).text() == "foobar"

    def test_completion(self, qtbot):
        widget = QSearchableListWidget()
        qtbot.addWidget(widget)
        widget.show()
        widget.addItems(["foo", "bar", "foobar", "baz", "bazbar", "foobaz"])
        widget.filter_widget.setText("fo")
        assert widget.count() == 6
        for i in range(widget.count()):
            item = widget.item(i)
            assert item.isHidden() == ("fo" not in item.text())

        widget.hide()



from contextlib import suppress
from platform import system

import pytest
from qtpy import QT_VERSION
from qtpy.QtCore import QEvent, QPoint, QPointF, Qt
from qtpy.QtGui import QHoverEvent, QMouseEvent, QWheelEvent

QT_VERSION = tuple(int(x) for x in QT_VERSION.split("."))

SYS_DARWIN = system() == "Darwin"

skip_on_linux_qt6 = pytest.mark.skipif(
    system() == "Linux" and QT_VERSION >= (6, 0),
    reason="hover events not working on linux pyqt6",
)

_PointF = QPointF()


def _mouse_event(pos=_PointF, type_=QEvent.Type.MouseMove):
    """Create a mouse event of `type_` at `pos`."""
    return QMouseEvent(
        type_,
        QPointF(pos),  # localPos
        QPointF(),  # windowPos / globalPos
        Qt.MouseButton.LeftButton,  # button
        Qt.MouseButton.LeftButton,  # buttons
        Qt.KeyboardModifier.NoModifier,  # modifiers
    )


def _wheel_event(arc):
    """Create a wheel event with `arc`."""
    with suppress(TypeError):
        return QWheelEvent(
            QPointF(),
            QPointF(),
            QPoint(arc, arc),
            QPoint(arc, arc),
            Qt.MouseButton.NoButton,
            Qt.KeyboardModifier.NoModifier,
            Qt.ScrollPhase.ScrollBegin,
            False,
            Qt.MouseEventSource.MouseEventSynthesizedByQt,
        )
    with suppress(TypeError):
        return QWheelEvent(
            QPointF(),
            QPointF(),
            QPoint(-arc, -arc),
            QPoint(-arc, -arc),
            1,
            Qt.Orientation.Vertical,
            Qt.MouseButton.NoButton,
            Qt.KeyboardModifier.NoModifier,
            Qt.ScrollPhase.ScrollBegin,
            False,
            Qt.MouseEventSource.MouseEventSynthesizedByQt,
        )

    return QWheelEvent(
        QPointF(),
        QPointF(),
        QPoint(arc, arc),
        QPoint(arc, arc),
        1,
        Qt.Orientation.Vertical,
        Qt.MouseButton.NoButton,
        Qt.KeyboardModifier.NoModifier,
    )


def _hover_event(_type, position, old_position, widget=None):
    with suppress(TypeError):
        return QHoverEvent(
            _type,
            position,
            widget.mapToGlobal(position),
            old_position,
        )
    return QHoverEvent(_type, position, old_position)


def _linspace(start: int, stop: int, n: int):
    h = (stop - start) / (n - 1)
    for i in range(n):
        yield start + h * i



from typing import Any, Iterable
from unittest.mock import Mock

import pytest

from superqt import QLabeledDoubleSlider, QLabeledRangeSlider, QLabeledSlider


def test_labeled_slider_api(qtbot):
    slider = QLabeledRangeSlider()
    qtbot.addWidget(slider)
    slider.hideBar()
    slider.showBar()
    slider.setBarVisible()
    slider.setBarMovesAllHandles()
    slider.setBarIsRigid()


def test_slider_connect_works(qtbot):
    slider = QLabeledSlider()
    qtbot.addWidget(slider)

    slider._label.editingFinished.emit()


def _assert_types(args: Iterable[Any], type_: type):
    # sourcery skip: comprehension-to-generator
    assert all(isinstance(v, type_) for v in args), "invalid type"


@pytest.mark.parametrize("cls", [QLabeledDoubleSlider, QLabeledSlider])
def test_labeled_signals(cls, qtbot):
    gslider = cls()
    qtbot.addWidget(gslider)

    type_ = float if cls == QLabeledDoubleSlider else int

    mock = Mock()
    gslider.valueChanged.connect(mock)
    with qtbot.waitSignal(gslider.valueChanged):
        gslider.setValue(10)
    mock.assert_called_once_with(10)
    _assert_types(mock.call_args.args, type_)

    mock = Mock()
    gslider.rangeChanged.connect(mock)
    with qtbot.waitSignal(gslider.rangeChanged):
        gslider.setMinimum(3)
    mock.assert_called_once_with(3, 99)
    _assert_types(mock.call_args.args, type_)

    mock.reset_mock()
    with qtbot.waitSignal(gslider.rangeChanged):
        gslider.setMaximum(15)
    mock.assert_called_once_with(3, 15)
    _assert_types(mock.call_args.args, type_)

    mock.reset_mock()
    with qtbot.waitSignal(gslider.rangeChanged):
        gslider.setRange(1, 2)
    mock.assert_called_once_with(1, 2)
    _assert_types(mock.call_args.args, type_)


@pytest.mark.parametrize(
    "cls", [QLabeledDoubleSlider, QLabeledRangeSlider, QLabeledSlider]
)
def test_editing_finished_signal(cls, qtbot):
    mock = Mock()
    slider = cls()
    qtbot.addWidget(slider)
    slider.editingFinished.connect(mock)
    if hasattr(slider, "_label"):
        slider._label.editingFinished.emit()
    else:
        slider._min_label.editingFinished.emit()
    mock.assert_called_once()


def test_editing_float(qtbot):
    slider = QLabeledDoubleSlider()
    qtbot.addWidget(slider)
    slider._label.setValue(0.5)
    slider._label.editingFinished.emit()
    assert slider.value() == 0.5



from qtpy.QtCore import QPoint, QRect, QSize, Qt
from qtpy.QtGui import QFontMetrics, QResizeEvent
from qtpy.QtWidgets import QLabel

from ._eliding import _GenericEliding


class QElidingLabel(_GenericEliding, QLabel):
    """
    A QLabel variant that will elide text (could add '…') to fit width.

    QElidingLabel()
    QElidingLabel(parent: Optional[QWidget], f: Qt.WindowFlags = ...)
    QElidingLabel(text: str, parent: Optional[QWidget] = None, f: Qt.WindowFlags = ...)

    For a multiline eliding label, use `setWordWrap(True)`.  In this case, text
    will wrap to fit the width, and only the last line will be elided.
    When `wordWrap()` is True, `sizeHint()` will return the size required to fit
    the full text.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if args and isinstance(args[0], str):
            self.setText(args[0])

    # Reimplemented _GenericEliding methods

    def setElideMode(self, mode: Qt.TextElideMode) -> None:
        """Set the elide mode to a Qt.TextElideMode."""
        super().setElideMode(mode)
        super().setText(self._elidedText())

    def setEllipsesWidth(self, width: int) -> None:
        """A width value to take into account ellipses width when eliding text.

        The value is deducted from the widget width when computing the elided version
        of the text.
        """
        super().setEllipsesWidth(width)
        super().setText(self._elidedText())

    # Reimplemented QT methods

    def text(self) -> str:
        """Return the label's text.

        If no text has been set this will return an empty string.
        """
        return self._text

    def setText(self, txt: str) -> None:
        """Set the label's text.

        Setting the text clears any previous content.
        NOTE: we set the QLabel private text to the elided version
        """
        self._text = txt
        super().setText(self._elidedText())

    def resizeEvent(self, event: QResizeEvent) -> None:
        event.accept()
        super().setText(self._elidedText())

    def setWordWrap(self, wrap: bool) -> None:
        super().setWordWrap(wrap)
        super().setText(self._elidedText())

    def sizeHint(self) -> QSize:
        if not self.wordWrap():
            return super().sizeHint()
        fm = QFontMetrics(self.font())
        flags = int(self.alignment() | Qt.TextFlag.TextWordWrap)
        r = fm.boundingRect(QRect(QPoint(0, 0), self.size()), flags, self._text)
        return QSize(self.width(), r.height())



from __future__ import annotations

__all__ = [
    "addFont",
    "Animation",
    "ENTRY_POINT",
    "font",
    "icon",
    "IconFont",
    "IconFontMeta",
    "IconOpts",
    "pulse",
    "setTextIcon",
    "spin",
]

from typing import TYPE_CHECKING

from ._animations import Animation, pulse, spin
from ._iconfont import IconFont, IconFontMeta
from ._plugins import FontIconManager as _FIM
from ._qfont_icon import DEFAULT_SCALING_FACTOR, IconOptionDict, IconOpts
from ._qfont_icon import QFontIconStore as _QFIS

if TYPE_CHECKING:
    from qtpy.QtGui import QFont, QTransform
    from qtpy.QtWidgets import QWidget

    from ._qfont_icon import QFontIcon, ValidColor

ENTRY_POINT = _FIM.ENTRY_POINT


# FIXME: currently, an Animation requires a *pre-bound* QObject.  which makes it very
# awkward to use animations when declaratively listing icons.  It would be much better
# to have a way to find the widget later, to execute the animation... short of that, I
# think we should take animation off of `icon` here, and suggest that it be an
# an additional convenience method after the icon has been bound to a QObject.
def icon(
    glyph_key: str,
    scale_factor: float = DEFAULT_SCALING_FACTOR,
    color: ValidColor | None = None,
    opacity: float = 1,
    animation: Animation | None = None,
    transform: QTransform | None = None,
    states: dict[str, IconOptionDict | IconOpts] | None = None,
) -> QFontIcon:
    """Create a QIcon for `glyph_key`, with a number of optional settings.

    The `glyph_key` (e.g. 'fa5s.smile') represents a Font-family & style, and a glyph.
    In most cases, the key should be provided by a plugin in the environment, like:

    - [fonticon-fontawesome5](https://pypi.org/project/fonticon-fontawesome5/) ('fa5s' &
      'fa5r' prefixes)
    - [fonticon-materialdesignicons6](https://pypi.org/project/fonticon-materialdesignicons6/)
      ('mdi6' prefix)

    ...but fonts can also be added manually using [`addFont`][superqt.fonticon.addFont].

    Parameters
    ----------
    glyph_key : str
        String encapsulating a font-family, style, and glyph. e.g. 'fa5s.smile'.
    scale_factor : float, optional
        Scale factor (fraction of widget height), When widget icon is painted on widget,
        it will use `font.setPixelSize(round(wdg.height() * scale_factor))`.
        by default 0.875.
    color : ValidColor, optional
        Color for the font, by default None. (e.g. The default `QColor`)
        Valid color types include `QColor`, `int`, `str`, `Qt.GlobalColor`, `tuple` (of
        integer: RGB[A]) (anything that can be passed to `QColor`).
    opacity : float, optional
        Opacity of icon, by default 1
    animation : Animation, optional
        Animation for the icon.  A subclass of superqt.fonticon.Animation, that provides
        a concrete `animate` method. (see "spin" and "pulse" for examples).
        by default None.
    transform : QTransform, optional
        A `QTransform` to apply when painting the icon, by default None
    states : dict, optional
        Provide additional styling for the icon in different states.  `states` must be
        a mapping of string to dict, where:

        - the key represents a `QIcon.State` ("on", "off"), a `QIcon.Mode` ("normal",
          "active", "selected", "disabled"), or any combination of a state & mode
          separated by an underscore (e.g. "off_active", "selected_on", etc...).
        - the value is a dict with all of the same key/value meanings listed above as
          parameters to this function (e.g. `glyph_key`, `color`,`scale_factor`,
          `animation`, etc...)

        Missing keys in the state dicts will be taken from the default options, provided
        by the parameters above.

    Returns
    -------
    QFontIcon
        A subclass of QIcon.  Can be used wherever QIcons are used, such as
        `widget.setIcon()`

    Examples
    --------
    simple example (using the string `'fa5s.smile'` assumes the `fonticon-fontawesome5`
    plugin is installed)

    >>> btn = QPushButton()
    >>> btn.setIcon(icon('fa5s.smile'))

    can also directly import from fonticon_fa5
    >>> from fonticon_fa5 import FA5S
    >>> btn.setIcon(icon(FA5S.smile))

    with animation
    >>> btn2 = QPushButton()
    >>> btn2.setIcon(icon(FA5S.spinner, animation=pulse(btn2)))

    complicated example
    >>> btn = QPushButton()
    >>> btn.setIcon(
    ...     icon(
    ...         FA5S.ambulance,
    ...         color="blue",
    ...         states={
    ...             "active": {
    ...                 "glyph": FA5S.bath,
    ...                 "color": "red",
    ...                 "scale_factor": 0.5,
    ...                 "animation": pulse(btn),
    ...             },
    ...             "disabled": {
    ...                 "color": "green",
    ...                 "scale_factor": 0.8,
    ...                 "animation": spin(btn)
    ...             },
    ...         },
    ...     )
    ... )
    >>> btn.setIconSize(QSize(256, 256))
    >>> btn.show()

    """
    return _QFIS.instance().icon(
        glyph_key,
        scale_factor=scale_factor,
        color=color,
        opacity=opacity,
        animation=animation,
        transform=transform,
        states=states or {},
    )


def setTextIcon(widget: QWidget, glyph_key: str, size: float | None = None) -> None:
    """Set text on a widget to a specific font & glyph.

    This is an alternative to setting a QIcon with a pixmap.  It may be easier to
    combine with dynamic stylesheets.

    Parameters
    ----------
    widget : QWidget
        A widget supporting a `setText` method.
    glyph_key : str
        String encapsulating a font-family, style, and glyph. e.g. 'fa5s.smile'.
    size : int, optional
        Size for QFont.  passed to `setPixelSize`, by default None
    """
    return _QFIS.instance().setTextIcon(widget, glyph_key, size)


def font(font_prefix: str, size: int | None = None) -> QFont:
    """Create QFont for `font_prefix`.

    Parameters
    ----------
    font_prefix : str
        Font_prefix, such as 'fa5s' or 'mdi6', representing a font-family and style.
    size : int, optional
        Size for QFont.  passed to `setPixelSize`, by default None

    Returns
    -------
    QFont
        QFont instance that can be used to add fonticons to widgets.
    """
    return _QFIS.instance().font(font_prefix, size)


def addFont(
    filepath: str, prefix: str, charmap: dict[str, str] | None = None
) -> tuple[str, str] | None:
    """Add OTF/TTF file at `filepath` to the registry under `prefix`.

    If you'd like to later use a fontkey in the form of `prefix.some-name`, then
    `charmap` must be provided and provide a mapping for all of the glyph names
    to their unicode numbers. If a charmap is not provided, glyphs must be directly
    accessed with their unicode as something like `key.\uffff`.

    !!! Note
        in most cases, users will not need this. Instead, they should install a
        font plugin, like:

        - [fonticon-fontawesome5](https://pypi.org/project/fonticon-fontawesome5/)
        - [fonticon-materialdesignicons6](https://pypi.org/project/fonticon-materialdesignicons6/)

    Parameters
    ----------
    filepath : str
        Path to an OTF or TTF file containing the fonts
    prefix : str
        A prefix that will represent this font file when used for lookup.  For example,
        'fa5s' for 'Font-Awesome 5 Solid'.
    charmap : Dict[str, str], optional
        optional mapping for all of the glyph names to their unicode numbers.
        See note above.

    Returns
    -------
    Tuple[str, str], optional
        font-family and font-style for the file just registered, or `None` if
        something goes wrong.
    """
    return _QFIS.instance().addFont(filepath, prefix, charmap)


del DEFAULT_SCALING_FACTOR
